#include "lexer.h"

#include "ctype.h"
#include "tokens.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static bool State_to_token_type(
    enum Lexer_State state,
    enum TokenType *token_type
);

static void token_start(Lexer lexer, enum Lexer_State state);
static void token_start_putc(Lexer lexer, enum Lexer_State state, char chr);
static void token_putc(Lexer lexer, char chr);
static void token_finish(Lexer lexer);

Lexer Lexer_new(const Tokens tokens)
{
    assert(tokens);

    const Lexer self = malloc(sizeof(struct Lexer));
    assert(self);

    memset(self, 0, sizeof(struct Lexer));
    self->tokens = tokens;
    self->state = STATE_INIT;
    self->buffer[0] = '\0';
    self->buffer_index = 0;

    return self;
}

void Lexer_delete(const Lexer self)
{
    assert(self);
    free(self);
}

bool
State_to_token_type(const enum Lexer_State state, enum TokenType *const token_type)
{
    switch (state) {
    case STATE_ROUND_OPEN:    *token_type = TOKEN_ROUND_OPEN;    break;
    case STATE_ROUND_CLOSE:   *token_type = TOKEN_ROUND_CLOSE;   break;
    case STATE_SQUARE_OPEN:   *token_type = TOKEN_SQUARE_OPEN;   break;
    case STATE_SQUARE_CLOSE:  *token_type = TOKEN_SQUARE_CLOSE;  break;
    case STATE_CURLY_OPEN:    *token_type = TOKEN_CURLY_OPEN;    break;
    case STATE_CURLY_CLOSE:   *token_type = TOKEN_CURLY_CLOSE;   break;
    case STATE_QUOTE:         *token_type = TOKEN_QUOTE;         break;
    case STATE_QUASI_QUOTE:   *token_type = TOKEN_QUASI_QUOTE;   break;
    case STATE_QUASI_UNQUOTE: *token_type = TOKEN_QUASI_UNQUOTE; break;
    case STATE_IDENT:         *token_type = TOKEN_IDENT;         break;
    case STATE_TAG:           *token_type = TOKEN_TAG;           break;
    case STATE_NUM:           *token_type = TOKEN_NUM;           break;
    case STATE_STRING_START:  *token_type = TOKEN_STRING;        break;
    case STATE_STRING_INSIDE: *token_type = TOKEN_STRING;        break;
    default: return false;
    }

    return true;
}

void token_start(const Lexer lexer, const enum Lexer_State state)
{
    assert(lexer);

    lexer->state = state;
    lexer->buffer_index = 0;
    lexer->buffer[0] = '\0';
}

void token_start_putc(
    const Lexer lexer,
    const enum Lexer_State state,
    const char chr
) {
    token_start(lexer, state);
    token_putc(lexer, chr);
}

void token_putc(const Lexer lexer, char chr)
{
    assert(lexer);
    assert(lexer->buffer_index < LEXER_BUFFER_SLEN);

    lexer->buffer[lexer->buffer_index] = chr;
    ++lexer->buffer_index;
    lexer->buffer[lexer->buffer_index] = '\0';
}

void token_finish(const Lexer lexer)
{
    assert(lexer);

    if (lexer->state != STATE_WHITESPACE &&
        lexer->state != STATE_COMMENT_LINE)
    {
        enum TokenType token_type;
        assert(State_to_token_type(lexer->state, &token_type));

        Tokens_append(lexer->tokens, token_type, lexer->buffer);
    }

    lexer->buffer_index = 0;
    lexer->buffer[0] = '\0';
}

void Lexer_lex(const Lexer self, const char chr)
{
    assert(self);
    assert(chr != EOF);

    switch (self->state) {
    case STATE_INIT:
        if (chr == ';') {
            token_start(self, STATE_COMMENT_LINE);
        } else if (chr == '(') {
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '#') {
            token_start(self, STATE_SHARP);
        } else if (chr == '"') {
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_start_putc(self, STATE_WHITESPACE, chr);
        } else if (is_ident_head(chr)) {
            token_start_putc(self, STATE_IDENT, chr);
        } else if (is_number(chr)) {
            token_start_putc(self, STATE_NUM, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_WHITESPACE:
        if (chr == ';') {
            token_start(self, STATE_COMMENT_LINE);
        } else if (chr == '(') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_finish(self);
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '#') {
            token_finish(self);
            token_start(self, STATE_SHARP);
        } else if (chr == '"') {
            token_finish(self);
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_putc(self, chr);
        } else if (is_ident_head(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_IDENT, chr);
        } else if (is_number(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_NUM, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_COMMENT_LINE:
        if (chr == '\n') {
            token_finish(self);
            token_start(self, STATE_WHITESPACE);
        } else {
            token_putc(self, chr);
        }
        break;
    case STATE_ROUND_OPEN:
    case STATE_ROUND_CLOSE:
    case STATE_SQUARE_OPEN:
    case STATE_SQUARE_CLOSE:
    case STATE_CURLY_OPEN:
    case STATE_CURLY_CLOSE:
    case STATE_QUOTE:
    case STATE_QUASI_QUOTE:
    case STATE_QUASI_UNQUOTE:
        if (chr == ';') {
            token_start(self, STATE_COMMENT_LINE);
        } else if (chr == '(') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_finish(self);
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '#') {
            token_finish(self);
            token_start(self, STATE_SHARP);
        } else if (chr == '"') {
            token_finish(self);
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_WHITESPACE, chr);
        } else if (is_ident_head(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_IDENT, chr);
        } else if (is_number(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_NUM, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_SHARP:
        if (is_tag(chr)) {
            token_start_putc(self, STATE_TAG, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_TAG:
        if (chr == ';') {
            self->state = STATE_COMMENT_LINE;
        } else if (chr == '(') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_finish(self);
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '"') {
            token_finish(self);
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_WHITESPACE, chr);
        } else if (is_tag(chr)) {
            token_putc(self, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_IDENT:
        if (chr == ';') {
            self->state = STATE_COMMENT_LINE;
        } else if (chr == '(') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_finish(self);
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '"') {
            token_finish(self);
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_WHITESPACE, chr);
        } else if (is_ident_tail(chr)) {
            token_putc(self, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_NUM:
        if (chr == ';') {
            token_start(self, STATE_COMMENT_LINE);
        } else if (chr == '(') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_finish(self);
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_finish(self);
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_finish(self);
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_finish(self);
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_finish(self);
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '"') {
            token_finish(self);
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_finish(self);
            token_start_putc(self, STATE_WHITESPACE, chr);
        } else if (is_number(chr)) {
            token_putc(self, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_STRING_START:
    case STATE_STRING_INSIDE:
        if (chr == '"') {
            token_finish(self);
            token_start(self, STATE_STRING_END);
        } else if (chr == '\\') {
            self->state = STATE_STRING_ESCAPE;
        } else {
            token_putc(self, chr);
        }
        break;
    case STATE_STRING_END:
        if (chr == ';') {
            token_start(self, STATE_COMMENT_LINE);
        } else if (chr == '(') {
            token_start_putc(self, STATE_ROUND_OPEN, chr);
        } else if (chr == ')') {
            token_start_putc(self, STATE_ROUND_CLOSE, chr);
        } else if (chr == '[') {
            token_start_putc(self, STATE_SQUARE_OPEN, chr);
        } else if (chr == ']') {
            token_start_putc(self, STATE_SQUARE_CLOSE, chr);
        } else if (chr == '{') {
            token_start_putc(self, STATE_CURLY_OPEN, chr);
        } else if (chr == '}') {
            token_start_putc(self, STATE_CURLY_CLOSE, chr);
        } else if (chr == '\'') {
            token_start_putc(self, STATE_QUOTE, chr);
        } else if (chr == '`') {
            token_start_putc(self, STATE_QUASI_QUOTE, chr);
        } else if (chr == ',') {
            token_start_putc(self, STATE_QUASI_UNQUOTE, chr);
        } else if (chr == '#') {
            token_start(self, STATE_SHARP);
        } else if (chr == '"') {
            token_start(self, STATE_STRING_START);
        } else if (is_space(chr)) {
            token_start_putc(self, STATE_WHITESPACE, chr);
        } else if (is_ident_head(chr)) {
            token_start_putc(self, STATE_IDENT, chr);
        } else if (is_number(chr)) {
            token_start_putc(self, STATE_NUM, chr);
        } else {
            assert(0);
        }
        break;
    case STATE_STRING_ESCAPE:
        if (chr == '\\' || chr == '"') {
            token_putc(self, chr);
            self->state = STATE_STRING_INSIDE;
        } else {
            assert(0);
        }
        break;
    }
}
