#ifndef __ARCANA_LISP_OBJECT_H__
#define __ARCANA_LISP_OBJECT_H__

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

enum Type {
    TYPE_PROCEDURE,
    TYPE_PAIR,
    TYPE_BOOLEAN,
    TYPE_CHAR,
    TYPE_SYMBOL,
    TYPE_STRING,
    TYPE_NUMBER,
};

/*************
 * Procedure *
 *************/

typedef struct Object *(*Procedure_Func)(
    size_t args_count,
    struct Object **args_array
);

struct Procedure {
    char *name;
    // For builtins
    Procedure_Func func;
    // For lambdas
    struct Object *arg_names;
    struct Object *body;
    struct Object *environment;
};

/********
 * Pair *
 ********/

struct Pair {
    struct Object *car, *cdr;
};

/**********
 * Number *
 **********/

struct Number {
    int64_t i64;
};

/**********
 * Object *
 **********/

struct Object {
    enum Type type;
    bool is_frozen;
    union {
        // For PROCEDURE
        struct Procedure procedure;
        // For PAIR
        struct Pair pair;
        // For BOOLEAN
        bool boolean;
        // For CHAR
        char chr;
        // For SYMBOL, STRING
        char *s;
        // For NUMBER
        struct Number number;
    };
};

const char *Type_to_str(enum Type type);

/************
 * Creation *
 ************/

struct Object *Object_new_pair(struct Object *car, struct Object *cdr);
struct Object *Object_new_boolean(bool boolean);
struct Object *Object_new_char(char chr);
struct Object *Object_new_symbol(const char *s);
struct Object *Object_new_string(const char *s);
struct Object *Object_new_number(int64_t i64);

struct Object *Object_new_lambda(
    const char *name,
    struct Object *arg_names,
    struct Object *body,
    struct Object *environment
);

struct Object *Object_build_list(int count, ...);

/*******************
 * Type predicates *
 *******************/

#define OBJECT_IS_NULL(object) ((object) == NULL)
bool Object_is_procedure(struct Object *object);
bool Object_is_pair(struct Object *object);
bool Object_is_boolean(struct Object *object);
bool Object_is_char(struct Object *object);
bool Object_is_symbol(struct Object *object);
bool Object_is_string(struct Object *object);
bool Object_is_number(struct Object *object);

/********************
 * Other predicates *
 ********************/

#define OBJECT_IS_LIST_HEAD(object) \
    (OBJECT_IS_NULL(object) || Object_is_pair(object))
bool Object_is_false(struct Object *object);
bool Object_is_true(struct Object *object);

/**************************************
 * Helper functions for specific type *
 **************************************/

size_t Object_list_length(struct Object *list_obj);
struct Object *Object_procedure_call(
    struct Object *procedure,
    struct Object *args
);

#endif
