#include "syntax.h"

#include "eval.h"
#include "object.h"

#include <assert.h>
#include <string.h>

/*******************************************
 * Special syntax to use in the executable *
 *******************************************/

struct Object *syntax_repl(
    struct Object *args,
    struct Object *const environment
) {
    return eval(args, environment);
}

struct Object *syntax_script(
    struct Object *args,
    struct Object *const environment
) {
    return syntax_begin(args, environment);
}

/*******************
 * Syntax keywords *
 *******************/

struct Object *syntax_begin(
    struct Object *args,
    struct Object *const environment
) {
    assert(OBJECT_IS_LIST_HEAD(args));
    struct Object *result = NULL;
    while (!OBJECT_IS_NULL(args)) {
        assert(Object_is_pair(args));
        result = eval(args->pair.car, environment);
        args = args->pair.cdr;
    }
    return result;
}

struct Object *syntax_cond(
    struct Object *const args,
    struct Object *const environment
) {
    assert(Object_is_pair(environment));
    assert(OBJECT_IS_LIST_HEAD(environment->pair.cdr));
    assert(OBJECT_IS_LIST_HEAD(environment->pair.car));

    assert(OBJECT_IS_LIST_HEAD(args));

    for (struct Object *arg = args; !OBJECT_IS_NULL(arg); arg = arg->pair.cdr)
    {
        assert(Object_is_pair(arg));
        assert(Object_is_pair(arg->pair.car));
        assert(Object_is_pair(arg->pair.car->pair.cdr));
        assert(OBJECT_IS_NULL(arg->pair.car->pair.cdr->pair.cdr));

        struct Object *const cond_prog = arg->pair.car->pair.car;

        if (Object_is_symbol(cond_prog) && strcmp(cond_prog->s, "else") == 0) {
            assert(OBJECT_IS_NULL(arg->pair.cdr));
        }
    }

    for (struct Object *arg = args; !OBJECT_IS_NULL(arg); arg = arg->pair.cdr)
    {
        struct Object *const cond_prog = arg->pair.car->pair.car;
        struct Object *const val_prog  = arg->pair.car->pair.cdr->pair.car;

        if (Object_is_symbol(cond_prog) && strcmp(cond_prog->s, "else") == 0) {
            return eval(val_prog, environment);
        }

        if (!Object_is_false(eval(cond_prog, environment))) {
            return eval(val_prog, environment);
        }
    }

    return NULL;
}

struct Object *syntax_define(
    struct Object *const args,
    struct Object *const environment
) {
    assert(Object_is_pair(environment));
    assert(OBJECT_IS_LIST_HEAD(environment->pair.cdr));
    assert(OBJECT_IS_LIST_HEAD(environment->pair.car));

    assert(Object_is_pair(args));
    struct Object *const name_and_args = args->pair.car;
    assert(Object_is_pair(args->pair.cdr));
    struct Object *const value_expr = args->pair.cdr->pair.car;
    assert(OBJECT_IS_NULL(args->pair.cdr->pair.cdr));

    struct Object *name  = NULL;
    struct Object *value = NULL;

    // Variable
    if (Object_is_symbol(name_and_args)) {
        name = name_and_args;
        value = eval(value_expr, environment);
    }
    // Lambda
    else {
        assert(Object_is_pair(name_and_args));

        name = name_and_args->pair.car;
        assert(Object_is_symbol(name));

        struct Object *const arg_names = name_and_args->pair.cdr;
        assert(OBJECT_IS_LIST_HEAD(arg_names));

        value = Object_new_lambda(name->s, arg_names, value_expr, environment);
    }

    environment->pair.car = Object_new_pair(
        Object_new_pair(name, value),
        environment->pair.car
    );

    return NULL;
}

struct Object *syntax_if(
    struct Object *const args,
    struct Object *const environment
) {
    assert(Object_is_pair(args));
    struct Object *const cond_expr = args->pair.car;
    struct Object *const then_else_list = args->pair.cdr;
    assert(then_else_list);
    assert(Object_is_pair(then_else_list));
    struct Object *const then_branch = then_else_list->pair.car;
    struct Object *const else_list = then_else_list->pair.cdr;
    assert(else_list);
    assert(Object_is_pair(else_list));
    struct Object *const else_branch = else_list->pair.car;
    assert(OBJECT_IS_NULL(else_list->pair.cdr));

    if (Object_is_false(eval(cond_expr, environment))) {
        return eval(else_branch, environment);
    } else {
        return eval(then_branch, environment);
    }
}

struct Object *syntax_lambda(
    struct Object *const args,
    struct Object *environment
) {
    assert(Object_is_pair(args));
    struct Object *const arg_names = args->pair.car;
    assert(OBJECT_IS_LIST_HEAD(arg_names));
    assert(Object_is_pair(args->pair.cdr));
    struct Object *const body = args->pair.cdr->pair.car;
    assert(OBJECT_IS_NULL(args->pair.cdr->pair.cdr));

    return Object_new_lambda(NULL, arg_names, body, environment);

}

struct Object *syntax_let(
    struct Object *const args,
    struct Object *environment
) {
    assert(Object_is_pair(args));
    struct Object *bindings = args->pair.car;
    assert(OBJECT_IS_LIST_HEAD(bindings));
    assert(Object_is_pair(args->pair.cdr));
    struct Object *const expr = args->pair.cdr->pair.car;
    assert(OBJECT_IS_NULL(args->pair.cdr->pair.cdr));

    assert(OBJECT_IS_LIST_HEAD(environment));
    environment = Object_new_pair(NULL, environment);

    while (!OBJECT_IS_NULL(bindings)) {
        assert(Object_is_pair(bindings));
        assert(Object_is_pair(bindings->pair.car));
        struct Object *const name = bindings->pair.car->pair.car;
        assert(Object_is_symbol(name));
        assert(Object_is_pair(bindings->pair.car->pair.cdr));
        struct Object *const val_expr = bindings->pair.car->pair.cdr->pair.car;
        assert(OBJECT_IS_NULL(bindings->pair.car->pair.cdr->pair.cdr));

        environment->pair.car = Object_new_pair(
            Object_new_pair(name, eval(val_expr, environment)),
            environment->pair.car
        );

        bindings = bindings->pair.cdr;
    }

    return eval(expr, environment);
}

struct Object *syntax_quote(
    struct Object *const args,
    struct Object *const environment
) {
    (void)environment; // unused

    assert(Object_is_pair(args));
    assert(OBJECT_IS_NULL(args->pair.cdr));
    return args->pair.car;
}
